@@ -9,19 +9,25 @@ from __future__ import absolute_import
 
 import collections
 import os
-import sys
+import threading
 import types
+from urllib.parse import urlparse
+from urllib.request import urlopen
+
 from ply import lex, yacc
+
+from ..thrift import TException, TPayload, TType, gen_init
+from .exc import ThriftGrammarError, ThriftParserError
 from .lexer import *  # noqa
-from .exc import ThriftParserError, ThriftGrammerError
-from thriftpy2._compat import urlopen, urlparse, PY3
-from ..thrift import gen_init, TType, TPayload, TException
+
+
+threadlocal = threading.local()
 
 
 def p_error(p):
     if p is None:
-        raise ThriftGrammerError('Grammer error at EOF')
-    raise ThriftGrammerError('Grammer error %r at line %d' %
+        raise ThriftGrammarError('Grammar error at EOF')
+    raise ThriftGrammarError('Grammar error %r at line %d' %
                              (p.value, p.lineno))
 
 
@@ -47,12 +53,12 @@ def p_header_unit(p):
 
 def p_include(p):
     '''include : INCLUDE LITERAL'''
-    thrift = thrift_stack[-1]
+    thrift = threadlocal.thrift_stack[-1]
     if thrift.__thrift_file__ is None:
         raise ThriftParserError('Unexpected include statement while loading '
                                 'from file like object.')
     replace_include_dirs = [os.path.dirname(thrift.__thrift_file__)] \
-        + include_dirs_
+        + threadlocal.include_dirs_
     for include_dir in replace_include_dirs:
         path = os.path.join(include_dir, p[2])
         if os.path.exists(path):
@@ -72,7 +78,7 @@ def p_namespace(p):
     '''namespace : NAMESPACE namespace_scope IDENTIFIER'''
     # namespace is useless in thriftpy2
     # if p[2] == 'py' or p[2] == '*':
-    #     setattr(thrift_stack[-1], '__name__', p[3])
+    #     setattr(threadlocal.thrift_stack[-1], '__name__', p[3])
 
 
 def p_namespace_scope(p):
@@ -104,15 +110,15 @@ def p_definition_unit(p):
 
 
 def p_const(p):
-    '''const : CONST field_type IDENTIFIER '=' const_value
-             | CONST field_type IDENTIFIER '=' const_value sep'''
+    '''const : CONST field_type IDENTIFIER '=' const_value type_annotations
+             | CONST field_type IDENTIFIER '=' const_value type_annotations sep'''
 
     try:
         val = _cast(p[2], p.lineno(3))(p[5])
     except AssertionError:
         raise ThriftParserError('Type error for constant %s at line %d' %
                                 (p[3], p.lineno(3)))
-    setattr(thrift_stack[-1], p[3], val)
+    setattr(threadlocal.thrift_stack[-1], p[3], val)
     _add_thrift_meta('consts', val)
 
 
@@ -158,12 +164,12 @@ def p_const_map_item(p):
 
 def p_const_ref(p):
     '''const_ref : IDENTIFIER'''
-    child = thrift_stack[-1]
+    child = threadlocal.thrift_stack[-1]
     for name in p[1].split('.'):
         father = child
         child = getattr(child, name, None)
         if child is None:
-            raise ThriftParserError('Cann\'t find name %r at line %d'
+            raise ThriftParserError('Can\'t find name %r at line %d'
                                     % (p[1], p.lineno(1)))
 
     if _get_ttype(child) is None or _get_ttype(father) == TType.I32:
@@ -185,13 +191,13 @@ def p_ttype(p):
 
 def p_typedef(p):
     '''typedef : TYPEDEF field_type IDENTIFIER type_annotations'''
-    setattr(thrift_stack[-1], p[3], p[2])
+    setattr(threadlocal.thrift_stack[-1], p[3], p[2])
 
 
 def p_enum(p):  # noqa
     '''enum : ENUM IDENTIFIER '{' enum_seq '}' type_annotations'''
     val = _make_enum(p[2], p[4])
-    setattr(thrift_stack[-1], p[2], val)
+    setattr(threadlocal.thrift_stack[-1], p[2], val)
     _add_thrift_meta('enums', val)
 
 
@@ -221,12 +227,12 @@ def p_struct(p):
 def p_seen_struct(p):
     '''seen_struct : STRUCT IDENTIFIER '''
     val = _make_empty_struct(p[2])
-    setattr(thrift_stack[-1], p[2], val)
+    setattr(threadlocal.thrift_stack[-1], p[2], val)
     p[0] = val
 
 
 def p_union(p):
-    '''union : seen_union '{' field_seq '}' '''
+    '''union : seen_union '{' field_seq '}' type_annotations'''
     val = _fill_in_struct(p[1], p[3])
     _add_thrift_meta('unions', val)
 
@@ -234,14 +240,14 @@ def p_union(p):
 def p_seen_union(p):
     '''seen_union : UNION IDENTIFIER '''
     val = _make_empty_struct(p[2])
-    setattr(thrift_stack[-1], p[2], val)
+    setattr(threadlocal.thrift_stack[-1], p[2], val)
     p[0] = val
 
 
 def p_exception(p):
     '''exception : EXCEPTION IDENTIFIER '{' field_seq '}' type_annotations '''
     val = _make_struct(p[2], p[4], base_cls=TException)
-    setattr(thrift_stack[-1], p[2], val)
+    setattr(threadlocal.thrift_stack[-1], p[2], val)
     _add_thrift_meta('exceptions', val)
 
 
@@ -249,7 +255,7 @@ def p_simple_service(p):
     '''simple_service : SERVICE IDENTIFIER '{' function_seq '}'
                 | SERVICE IDENTIFIER EXTENDS IDENTIFIER '{' function_seq '}'
     '''
-    thrift = thrift_stack[-1]
+    thrift = threadlocal.thrift_stack[-1]
 
     if len(p) == 8:
         extends = thrift
@@ -384,21 +390,30 @@ class CurrentIncompleteType(dict):
         return self.index + 1
 
 
-incomplete_type = CurrentIncompleteType()
-
-
 def p_ref_type(p):
     '''ref_type : IDENTIFIER'''
-    ref_type = thrift_stack[-1]
+    ref_type = threadlocal.thrift_stack[-1]
 
-    for index, name in enumerate(p[1].split('.')):
-        ref_type = getattr(ref_type, name, None)
-        if ref_type is None:
-            if index != len(p[1].split('.')) - 1:
-                raise ThriftParserError('No type found: %r, at line %d' %
-                                        (p[1], p.lineno(1)))
-            p[0] = incomplete_type.set_info((p[1], p.lineno(1)))
-            return
+    for attr in dir(ref_type):
+        if attr in {'__doc__', '__loader__', '__name__', '__package__',
+                    '__spec__', '__thrift_file__', '__thrift_meta__'}:
+            continue
+        if p[1].startswith(attr + '.'):
+            name = p[1][len(attr) + 1:]
+            included_ref_type = getattr(ref_type, attr)
+            resolved_ref_type = getattr(included_ref_type, name, None)
+            if resolved_ref_type is not None:
+                ref_type = resolved_ref_type
+                break
+    else:
+        for index, name in enumerate(p[1].split('.')):
+            ref_type = getattr(ref_type, name, None)
+            if ref_type is None:
+                if index != len(p[1].split('.')) - 1:
+                    raise ThriftParserError('No type found: %r, at line %d' %
+                                            (p[1], p.lineno(1)))
+                p[0] = threadlocal.incomplete_type.set_info((p[1], p.lineno(1)))
+                return
 
     if hasattr(ref_type, '_ttype'):
         p[0] = getattr(ref_type, '_ttype'), ref_type
@@ -497,11 +512,6 @@ def p_type_annotation(p):
         p[0] = p[1], None  # Without Value
 
 
-thrift_stack = []
-include_dirs_ = ['.']
-thrift_cache = {}
-
-
 def parse(path, module_name=None, include_dirs=None, include_dir=None,
           lexer=None, parser=None, enable_cache=True, encoding='utf-8'):
     """Parse a single thrift file to module object, e.g.::
@@ -517,7 +527,7 @@ def parse(path, module_name=None, include_dirs=None, include_dir=None,
                          the `include` directive, by default: ['.'].
     :param include_dir: directory to find child thrift files. Note this keyword
                         parameter will be deprecated in the future, it exists
-                        for compatiable reason. If it's provided (not `None`),
+                        for compatible reason. If it's provided (not `None`),
                         it will be appended to `include_dirs`.
     :param lexer: ply lexer to use, if not provided, `parse` will new one.
     :param parser: ply parser to use, if not provided, `parse` will new one.
@@ -525,33 +535,35 @@ def parse(path, module_name=None, include_dirs=None, include_dir=None,
                          cached, this is enabled by default. If `module_name`
                          is provided, use it as cache key, else use the `path`.
     """
-    if os.name == 'nt' and sys.version_info[0] < 3:
-        os.path.samefile = lambda f1, f2: os.stat(f1) == os.stat(f2)
+    # threadlocal should be initialized in every threads
+    initialized = getattr(threadlocal, 'initialized', None)
+    if initialized is None:
+        threadlocal.thrift_stack = []
+        threadlocal.include_dirs_ = ['.']
+        threadlocal.thrift_cache = {}
+        threadlocal.incomplete_type = CurrentIncompleteType()
+        threadlocal.initialized = True
 
     # dead include checking on current stack
-    for thrift in thrift_stack:
+    for thrift in threadlocal.thrift_stack:
         if thrift.__thrift_file__ is not None and \
                 os.path.samefile(path, thrift.__thrift_file__):
             raise ThriftParserError('Dead including on %s' % path)
 
-    global thrift_cache
-
     cache_key = module_name or os.path.normpath(path)
 
-    if enable_cache and cache_key in thrift_cache:
-        return thrift_cache[cache_key]
+    if enable_cache and cache_key in threadlocal.thrift_cache:
+        return threadlocal.thrift_cache[cache_key]
 
     if lexer is None:
         lexer = lex.lex()
     if parser is None:
         parser = yacc.yacc(debug=False, write_tables=0)
 
-    global include_dirs_
-
     if include_dirs is not None:
-        include_dirs_ = include_dirs
+        threadlocal.include_dirs_ = include_dirs
     if include_dir is not None:
-        include_dirs_.append(include_dir)
+        threadlocal.include_dirs_.append(include_dir)
 
     if not path.endswith('.thrift'):
         raise ThriftParserError('Path should end with .thrift')
@@ -561,7 +573,7 @@ def parse(path, module_name=None, include_dirs=None, include_dir=None,
         with open(urlparse(path).netloc + urlparse(path).path) as fh:
             data = fh.read()
     elif len(url_scheme) <= 1:
-        with open(path) as fh:
+        with open(path, encoding=encoding) as fh:
             data = fh.read()
     elif url_scheme in ('http', 'https'):
         data = urlopen(path).read()
@@ -570,7 +582,7 @@ def parse(path, module_name=None, include_dirs=None, include_dir=None,
                                 'with path in protocol \'{}\''.format(
                                     url_scheme))
 
-    if PY3 and isinstance(data, bytes):
+    if isinstance(data, bytes):
         data = data.decode(encoding)
 
     if module_name is not None and not module_name.endswith('_thrift'):
@@ -583,13 +595,13 @@ def parse(path, module_name=None, include_dirs=None, include_dir=None,
 
     thrift = types.ModuleType(module_name)
     setattr(thrift, '__thrift_file__', path)
-    thrift_stack.append(thrift)
+    threadlocal.thrift_stack.append(thrift)
     lexer.lineno = 1
     parser.parse(data)
-    thrift_stack.pop()
+    threadlocal.thrift_stack.pop()
 
     if enable_cache:
-        thrift_cache[cache_key] = thrift
+        threadlocal.thrift_cache[cache_key] = thrift
     return thrift
 
 
@@ -602,19 +614,28 @@ def parse_fp(source, module_name, lexer=None, parser=None, enable_cache=True):
         <module 'note_thrift' (built-in)>
 
     :param source: file-like object, expected to have a method named `read`.
-    :param module_name: the name for parsed module, shoule be endswith
+    :param module_name: the name for parsed module, should be endswith
                         '_thrift'.
     :param lexer: ply lexer to use, if not provided, `parse` will new one.
     :param parser: ply parser to use, if not provided, `parse` will new one.
     :param enable_cache: if this is set to be `True`, parsed module will be
                          cached by `module_name`, this is enabled by default.
     """
+    # threadlocal should be initialized in every threads
+    initialized = getattr(threadlocal, 'initialized', None)
+    if initialized is None:
+        threadlocal.thrift_stack = []
+        threadlocal.include_dirs_ = ['.']
+        threadlocal.thrift_cache = {}
+        threadlocal.incomplete_type = CurrentIncompleteType()
+        threadlocal.initialized = True
+
     if not module_name.endswith('_thrift'):
         raise ThriftParserError('thriftpy2 can only generate module with '
                                 '\'_thrift\' suffix')
 
-    if enable_cache and module_name in thrift_cache:
-        return thrift_cache[module_name]
+    if enable_cache and module_name in threadlocal.thrift_cache:
+        return threadlocal.thrift_cache[module_name]
 
     if not hasattr(source, 'read'):
         raise ThriftParserError('Expected `source` to be a file-like object '
@@ -629,18 +650,18 @@ def parse_fp(source, module_name, lexer=None, parser=None, enable_cache=True):
 
     thrift = types.ModuleType(module_name)
     setattr(thrift, '__thrift_file__', None)
-    thrift_stack.append(thrift)
+    threadlocal.thrift_stack.append(thrift)
     lexer.lineno = 1
     parser.parse(data)
-    thrift_stack.pop()
+    threadlocal.thrift_stack.pop()
 
     if enable_cache:
-        thrift_cache[module_name] = thrift
+        threadlocal.thrift_cache[module_name] = thrift
     return thrift
 
 
 def _add_thrift_meta(key, val):
-    thrift = thrift_stack[-1]
+    thrift = threadlocal.thrift_stack[-1]
 
     if not hasattr(thrift, '__thrift_meta__'):
         meta = collections.defaultdict(list)
@@ -648,6 +669,10 @@ def _add_thrift_meta(key, val):
     else:
         meta = getattr(thrift, '__thrift_meta__')
 
+    if key != 'consts' and val.__name__ in [x.__name__ for x in meta[key]]:
+        raise ThriftGrammarError(('\'%s\' type is already defined in '
+                                  '\'%s\'') % (val.__name__, key))
+
     meta[key].append(val)
 
 
@@ -812,7 +837,10 @@ def _cast_struct(t):   # struct/exception/union
 
 
 def _make_enum(name, kvs):
-    attrs = {'__module__': thrift_stack[-1].__name__, '_ttype': TType.I32}
+    attrs = {
+        '__module__': threadlocal.thrift_stack[-1].__name__,
+        '_ttype': TType.I32
+    }
     cls = type(name, (object, ), attrs)
 
     _values_to_names = {}
@@ -836,7 +864,10 @@ def _make_enum(name, kvs):
 
 
 def _make_empty_struct(name, ttype=TType.STRUCT, base_cls=TPayload):
-    attrs = {'__module__': thrift_stack[-1].__name__, '_ttype': ttype}
+    attrs = {
+        '__module__': threadlocal.thrift_stack[-1].__name__,
+        '_ttype': ttype
+    }
     return type(name, (base_cls, ), attrs)
 
 
@@ -847,7 +878,7 @@ def _fill_in_struct(cls, fields, _gen_init=True):
 
     for field in fields:
         if field[0] in thrift_spec or field[3] in _tspec:
-            raise ThriftGrammerError(('\'%d:%s\' field identifier/name has '
+            raise ThriftGrammarError(('\'%d:%s\' field identifier/name has '
                                       'already been used') % (field[0],
                                                               field[3]))
         ttype = field[2]
@@ -872,12 +903,16 @@ def _make_service(name, funcs, extends):
     if extends is None:
         extends = object
 
-    attrs = {'__module__': thrift_stack[-1].__name__}
+    attrs = {'__module__': threadlocal.thrift_stack[-1].__name__}
     cls = type(name, (extends, ), attrs)
     thrift_services = []
 
     for func in funcs:
         func_name = func[2]
+        if func_name in thrift_services:
+            raise ThriftGrammarError(('\'%s\' function is already defined in '
+                                      'service \'%s\'') % (func_name,
+                                                           name))
         # args payload cls
         args_name = '%s_args' % func_name
         args_fields = func[3]
